
#include<iostream>
#include<fstream>

using namespace std;

#include"defo.h"

#include"vfkso.h"
#include"multiint.h"
#include"rigkserr.h"
#include"zgfiles.h"
#include"galks.h"


////////////// GALERKIN PROJECTION for KS

void mid(interval m[],int pdim)
{
  for(int i=1; i<= pdim; i++)
   m[i]=mid(m[i]);
}

interval volume(const interval x[],int dim)
// returns the volume of x
{
  interval ret=1.0;
  for(int i=1;i<=dim;i++){
    ret*=(Right(x[i])-Left(x[i]));
  }
  return ret;
}


int setcontainedinboxes(int dim,const interval enclx[],int boxnum,intervalmatrix1 *pboxes)
  //  boxnum - number of boxes s
  //  returns 1 if enclx=\sum_{i=1,..,boxnum} pboxes[i]
  // all objects are dim-dimensional
  //  0 - is returned when the function fails to establish the inclusion,
  //    but is still possible the inclusion is true
{
  int found=0;
  int i=1;
  while((i<=boxnum) && (!found)){
    found=MultiIn(enclx,pboxes[i],0,dim);
    i++;
  }
  int ret=found;

  if(!found){
    // I will see if <enclx> is the sum by suitably dividing it

    double volencl=abs(volume(enclx,dim)).right_b();
    interval center_enx[MAX_DIM+1];
    copyvector(enclx,center_enx,dim);
    mid(center_enx,dim);

    //  finding box <pboxes[i]> such that center_enx \in <pboxes[i[>
    //  the the overlap is largest possible
    int selbox=0;
    double overlap=0;
    double tempoverlap;
    interval temp[MAX_DIM+1];
    for(int i=1;i<= boxnum;i++){
      if(MultiIn(center_enx,pboxes[i],1,dim)){
        commonpart(enclx,pboxes[i],temp,dim); // dim= enclx \cap pboxes[i]
        tempoverlap=abs(volume(temp,dim)).right_b()/volencl;
        if(selbox >0){
          if(tempoverlap > overlap){
            selbox=i;
            overlap=tempoverlap;
          }
        } else{
          selbox=i;
          overlap=tempoverlap;
        }
      }
    }

    if(selbox!=0){ // we have found the 'best' box (overlapwise)
      interval part_enclx[MAX_DIM+1];
      int covered=1;
      int i=1;
      while((i<=dim) && covered){
        if(enclx[i].left_b() < (pboxes[selbox][i]).left_b()){
          copyvector(enclx,part_enclx,dim);
          part_enclx[i]=interval(enclx[i].left_b(),(pboxes[selbox][i]).left_b());

          int cbox=1;
          int found=0;
          while((cbox<=boxnum) && (!found)){
            found=MultiIn(enclx,pboxes[cbox],0,dim);
            cbox++;
          }
          covered=found;
        }

        if(covered){
          // now the right end of enclx[i]
          if(enclx[i].right_b() > (pboxes[selbox][i]).right_b()){
            copyvector(enclx,part_enclx,dim);
            part_enclx[i]=interval((pboxes[selbox][i]).right_b(),enclx[i].right_b());

            int cbox=1;
            int found=0;
            while((cbox<=boxnum) && (!found)){
              found=MultiIn(enclx,pboxes[cbox],0,dim);
              cbox++;
            }
            covered=found;
          }
        }
        i++;
      }
      if(covered) ret=1;
    }
  }
  return ret;
}





int Galerkin_Proj_KuraShi::fullcheckstatus(int newstatus)
    // sets new fullcheckstatus, returns previous value
{
  int ret=fullenclcheck;
  if(!newstatus && tailevolution && !fixedtaildefined){
    cerr << " Galerkin_Proj_KuraShi::fullcheckstatus  - we don't have tail "
         << endl;
    cerr << "can't change <fullenclstatus> to 0 " << endl;
     #if defined(KRAK) && defined(WIN95)
                   waitBt();
    #endif
    exit(EXIT_FAILURE);
  }
  fullenclcheck=newstatus;
  return ret;
}

int Galerkin_Proj_KuraShi::fullcheckstatus()
{
  return fullenclcheck;
}



int Galerkin_Proj_KuraShi::get_tailevolution()
{
  return tailevolution;
}


Galerkin_Proj_KuraShi::Galerkin_Proj_KuraShi(KuraShi *kurashi,int proj_dim,
          int pLdelta_method,int pLdelta_norm)
  // the constructor with tail evolution, the dimension of the projection
  // is given by <proj_dim>, kurashi->dim is set to <proj_dim>
{
  pode=kurashi;
  pkurashi=kurashi;

  strcpy(infostr,pkurashi->infostr);
  dim=proj_dim;
  pode->dim=proj_dim;
  dissipative=pkurashi->dissipative;

  for(int i=1;i<=dim;i++){
    y_c[i]=0;
    y_c_double[i]=0;
  }


  zero(inclusion_interval,dim);

  fixedtaildefined=0;
  pks_tail=NULL;
  gl_galerr=NULL;
  loc_galerr=NULL;


  uselocal_galerr=0;
  tailevolution=1;
  fullenclcheck=1;

  enc_order=pode->enc_order;

  divsecreg=pode->divsecreg;
  lognormused=pode->lognormused;

  Ldelta_method=pLdelta_method;

  if(pLdelta_norm!=0)
    Ldelta_norm=pLdelta_norm;
  else
  Ldelta_norm=NORM_MAX;

  boxnum=0;
  pboxes=NULL;

}




Galerkin_Proj_KuraShi::Galerkin_Proj_KuraShi(KuraShi *kurashi,
  interval galerr[],int proj_dim,int pLdelta_method,int pLdelta_norm,
        intervalmatrix1 *boxes,int par_boxnum,KuraShi_Tail *kstail,
        int puselocgalerr,int ptailevolution)
{
  if(puselocgalerr && ptailevolution){
      cerr << "Error - in constructor of Galerkin_Proj_KuraShi " << endl;
      #if defined(WIN95) && defined(KRAK)
        cerr << "press any key " << endl;
        waitBt();
      #endif


    exit(EXIT_FAILURE);
  }
  assignvalues(kurashi);

  pode=kurashi;

  pkurashi=kurashi;
  pks_tail=kstail;

  strcpy(infostr,pode->infostr);
  dim=proj_dim;
  pode->dim=proj_dim;

  for(int i=1;i<=dim;i++){
    y_c[i]=0;
    y_c_double[i]=0;
  }

  fixedtaildefined=0;
  copyvector(galerr,inclusion_interval,dim);

  gl_galerr=new interval[dim+1];
  loc_galerr=new interval[dim+1];
  copyvector(galerr,gl_galerr,dim);
  copyvector(galerr,loc_galerr,dim);
  uselocal_galerr=puselocgalerr;
  tailevolution=ptailevolution;



  Ldelta_method=pLdelta_method; //      the method used in LDelta computation
//  Ldelta_method=Ldelta_lognorm; // based on logarithmic norm
// Ldelta_method= Ldelta_componentwiseestm; // based on component-wise estimates


  if(pLdelta_norm!=0)
    Ldelta_norm=pLdelta_norm;
  else
  Ldelta_norm=NORM_MAX;
//  Ldelta_norm=NORM_SUM;

  boxnum=par_boxnum;
  pboxes=boxes;

  fullenclcheck=0;
  if(tailevolution) fullenclcheck=1;
}

Galerkin_Proj_KuraShi::~Galerkin_Proj_KuraShi()
{
  delete [] gl_galerr;
  delete [] loc_galerr;
}


interval Galerkin_Proj_KuraShi::norm(int nonum,interval x[],KuraShi_Tail *T)
//  nonum=1,2,3,4 - correspond respectively to l2,h1,c0 and c1-norms
{
  interval* ta;
  ta= new interval[T->big_m+1];

  for(int i=1;i<=T->m;i++)
    ta[i]=x[i];
  for(int i=T->m+1; i<=T->big_m;i++)
    ta[i]=(T->a)[i];

  interval ret=0;
  switch(nonum){
    case 1:  // l2norm
      ret=ksl2norm(0,ta,T->big_m,T->c,T->s);
      break;
    case 2: // h1norm
      ret=ksl2norm(1,ta,T->big_m,T->c,T->s);
      break;
    case 3: // c0-norm
      ret=ksuniformnorm(0,ta,T->big_m,T->c,T->s);
      break;
    case 4: // c1-norm
      ret=ksuniformnorm(1,ta,T->big_m,T->c,T->s);
      break;
  }

  delete [] ta;
  return ret;
}



interval  Galerkin_Proj_KuraShi::l2norm(interval x[],SetWithTail &y)
{
  return norm(1,x,y.ptail);
}

interval  Galerkin_Proj_KuraShi::h1norm(interval x[],SetWithTail &y)
{
  return norm(2,x,y.ptail);
}

interval  Galerkin_Proj_KuraShi::c0norm(interval x[],SetWithTail &y)
{
  return norm(3,x,y.ptail);
}

interval  Galerkin_Proj_KuraShi::c1norm(interval x[],SetWithTail &y)
{
  return norm(4,x,y.ptail);
}


void Galerkin_Proj_KuraShi::compute_loc_galerr(interval x[])
 // computes Galerkin errors for x \in x[] and the tail pks_tail
 // the result is placed in this->loc_galerr
{
  interval *q;
  q=new interval[pks_tail->big_m+1];

  copyvector(x,q,dim);  // dim=pkstail->m
  for(int i=dim+1;i<=pks_tail->big_m;i++)
    q[i]=(pks_tail->a)[i];

  ksgalerr_point(q,pks_tail->c,pks_tail->s,pks_tail->m,pks_tail->big_m,loc_galerr);
  delete [] q;
}





int Galerkin_Proj_KuraShi::tevol_incl_enclosure(const interval x[],SetWithTail &yin,
     interval h0,interval enclx[],SetWithTail &encly, int nomessages/* =0*/)
  // the enclosure function with the tail evolution
{
  int ret;
  // it may happen that encly - does not have its own tail defined
  KuraShi_Tail Tencl;

  #if !defined(ENCL_TEST)
  ret=pkurashi->enclosure_with_tail(h0,x,*(yin.ptail),enclx,Tencl,0,
       !fullenclcheck,gl_galerr,nomessages);
  #else
  cerr << "undefine ENCL_TEST in vfkso.h" << endl;
  #if defined(KRAK) && defined(WIN95)
  waitBt();
  #endif
  exit(EXIT_FAILURE);
  #endif

  if(ret){
    encly.assigntail(Tencl);
    pkurashi->galerkin_errors(x,*(encly.ptail),encly.y);
    if(fullenclcheck)
      copyvector(encly.y,inclusion_interval,pkurashi->dim);
        // this is the place where we update <inclusion_interval>
        // to be used during the entire time step
    else{
      if(!commonpart(encly.y,gl_galerr,inclusion_interval,pkurashi->dim)){
        cerr << "Gale.._KuraShi::incl_enclosure - an empty intersection "
            " of local and ariori gal. err" << endl;
        #if defined(WIN95) && defined(KRAK)
        cerr << "press any key" << endl;
        waitBt();
        #endif
        exit(EXIT_FAILURE);
      }
      copyvector(inclusion_interval,encly.y,pkurashi->dim);
    }
  }
  return ret;
}


int Galerkin_Proj_KuraShi::fixedtail_incl_enclosure(const interval x[],
     SetWithTail &yin,
     interval h0,interval enclx[],SetWithTail &encly, int nomessages/* =0*/)
{
  if(uselocal_galerr)
    copyvector(gl_galerr,inclusion_interval,dim);  // inclusion_interval - always contains the current
                // galerkin error - hence this is like the initialization
               // of <inclusion_interval> for the current time step,

  int ret=this->Perturbed_Ode::incl_enclosure(x,yin,h0,enclx,encly,nomessages);
  if(ret && uselocal_galerr){
    // we take an intersection of global bounds for gal. error (a priori)
    //  with local bounds ; obviously if <enclx> was in the range over which
    //  global bounds were defined, then we will have that local bounds are
    //  contained in global bounds; but the preliminary rounds (experiments
    //           with !fullenclcheck) this is not always true

    int iter=1;
    int maxiter=1; // maxiter=2  causes almost no improvement for atracting fixed
                 // points around mu=0.127
    int done=0;
    while(!done && (iter <= maxiter) && ret){
      compute_loc_galerr(enclx);  // the computed galerkin errors are placed in this->loc_galerr
      commonpart(loc_galerr,inclusion_interval,inclusion_interval,dim);
      ret=this->Perturbed_Ode::incl_enclosure(x,yin,h0,enclx,encly,nomessages);
      iter++;
    }
  }

  if(ret){
    if(fullenclcheck){
      ret=check_apriori_bounds(x,h0,enclx,nomessages);
    }
  }
  return ret;
}


int Galerkin_Proj_KuraShi::incl_enclosure(const interval x[],SetWithTail &yin,
     interval h0,interval enclx[],SetWithTail &encly, int nomessages/* =0*/)
  // if true is returned then:
   //    \varphi([0,h0],(x,yin)) \subset enclx \times encly
   //   and  inclusion_interval=galerkin errors for this time step
   //
   // if nomessages==0 - then  when procedure fails then a message is send
   //             to a screen and files
   // if nomessages==1 -  the failure does not result in message to screen
{

 int ret;
 if(!tailevolution)
     ret=fixedtail_incl_enclosure(x,yin,h0,enclx,encly,
            nomessages);
 else
     ret=tevol_incl_enclosure(x,yin,h0,enclx,encly,nomessages);
 return ret;
}


int Galerkin_Proj_KuraShi::y_move(interval xencl[],SetWithTail & yencl,interval h,
          SetWithTail & y,SetWithTail &yn)
    //  y - is mapped to yn after time step h; with xencl and yecnl
    // being enclosures for x and y, respectively
{
  int ret;
  if(!tailevolution)
    ret=this->Perturbed_Ode::y_move(xencl,yencl,h,y,yn);
  else{
   // it may happen that encly - does not have its own tail defined
    KuraShi_Tail Tn;
    ret=pkurashi->move_tail(h,xencl,*(y.ptail),*(yencl.ptail),Tn);
    if(ret){
      yn.assigntail(Tn);
      copyvector(yencl.y,yn.y,pkurashi->dim);
    }else{
      cerr << "problem with validation in y_move " << endl;
    }
  }
  return ret;
}

void checkinclusion_rep(ostream &report,const interval x[],const interval boxeshull[],
    int pdim)
// writes down to the file the 'margins' for the inclusion x \subset boxeshull
{
  for(int i=1;i<=pdim;i++){
    double rmargin=boxeshull[i].right_b() - x[i].right_b();
    double lmargin=x[i].left_b() - boxeshull[i].left_b();
    report << "i=" << i << "   " ;
    if( (rmargin >0) && (lmargin >0) ) {
      report << "ok  ";
      report << "margins " << lmargin << "   " << rmargin <<endl;
    } else{
      report << "not contained - margins  = " << lmargin << "  " << rmargin << endl;
    }
  }
}

void findintersectingboxes(int pdim,const interval x[],const interval boxes[][MAX_DIM+1],
   int boxnum,int tab_nonemptyints[],int *num_nonemptyints)
//  On output : tab_nomemptyints[1...*numnonemptyints] contains
//          indices, i, to boxes[], such that
//    boxes[i] has a nonempty intersection with x
// /////////////
//    <pdim>=dim(x)=dim(boxes[i]),   <boxnum> - the number of boxes in
//             <boxes>
{
  int count=0;
  interval cp[MAX_DIM+1];
  for(int i=1;i<=boxnum;i++){
    if(commonpart(x,boxes[i],cp,pdim)){
      count++;
      tab_nonemptyints[count]=i;
    }
  }
  *num_nonemptyints=count;
}


void listboxes(ostream &report,int /*pdim*/,const interval /*x*/[],
              const interval /*boxes*/[][MAX_DIM+1],const int tab_x_nonemptyints[],
              int box_nempty_x)
{
  report << box_nempty_x << " boxes with nonempty intersection " << endl;
  for(int i=1;i<=box_nempty_x;i++)
    report << tab_x_nonemptyints[i] << "  ";

  report << endl;

}


void Galerkin_Proj_KuraShi::diag_apriori_bounds(const interval x[],interval h0,const interval enclx[])
// the function writes to the file <zglerrorlogname>
// the data from failed enclosure procedure
///
// it tries to give a reason for this
{
  ofstream report(zglerrorlogname,ios::app);
  entriesinerrorlog=1;
  report << "-------------------------" << endl;
  writeCompInfo(report);
  writeparameters(report);
  writedatetime(report);
  report << "failed check of apriori bounds for differential inclusion" << endl;


  report << "h=" << h0 << endl ;

  report << "x" << endl;
  saveAllCoord(report,x,dim,2);
  report << endl << endl;


  report << "enclx" << endl;
  saveAllCoord(report,enclx,dim,2);
  report << endl << endl;



  int *tab_x_nemp_ints;
  tab_x_nemp_ints=new int[boxnum+1];
  int *tab_encl_nemp_ints;
  tab_encl_nemp_ints=new int[boxnum+1];

  int box_nemp_x;
  int box_nemp_encl;

  interval boxeshull[MAX_DIM+1];
  copyvector(pboxes[1],boxeshull,dim);
  for(int i=2;i<=boxnum;i++)
    MultiIhull(pboxes[i],boxeshull,boxeshull,dim);


  report << " A check of initial condition with a global hull " << endl;
  checkinclusion_rep(report,x,boxeshull,dim);


  report << endl << endl;
  report << "Check of enclx with a global hull " << endl;
  checkinclusion_rep(report,enclx,boxeshull,dim);


  findintersectingboxes(dim,x,pboxes,boxnum,tab_x_nemp_ints,&box_nemp_x);
  findintersectingboxes(dim,enclx,pboxes,boxnum,tab_encl_nemp_ints,&box_nemp_encl);

  report << endl;
  report << "Boxes intersecting x " << endl;
  listboxes(report,dim,x,pboxes,tab_x_nemp_ints,box_nemp_x);
  report << endl;

  report << "An inclusion check for x " << endl;
  for(int i=1; i<= box_nemp_x; i++){
    int bno=tab_x_nemp_ints[i];
    report << "box " << bno << endl;
    checkinclusion_rep(report,x,pboxes[bno],dim);
  }


  report << "Boxes intersecting enclx " << endl;
  listboxes(report,dim,enclx,pboxes,tab_encl_nemp_ints,box_nemp_encl);
  report << "An inclusion check for enclx " << endl;
  for(int i=1; i<= box_nemp_encl; i++){
    int bno=tab_encl_nemp_ints[i];
    report << "box " << bno << endl;
    checkinclusion_rep(report,enclx,pboxes[bno],dim);
  }
  delete [] tab_x_nemp_ints;
  delete [] tab_encl_nemp_ints;

  cout << " see the report file on bad enclosure " << zglerrorlogname << endl;

}


int  Galerkin_Proj_KuraShi::check_apriori_bounds(const interval x[],interval h0,
   interval enclx[],int nomessages /*=0*/)
  // 1 if returned if enclx \subset boxes
  // if nomessages==0 , then  when procedure fails then a message is send
   //             to a screen and files
   // if nomessages=1 , the the failure does not result in any error reports
{

  int ret=setcontainedinboxes(dim,enclx,boxnum,pboxes);

  if(!ret && !nomessages){
    cerr << "the enclosure with full check of apriori bounds went bad" << endl;
    diag_apriori_bounds(x,h0,enclx);
    #ifdef WIN95
//      cerr << "press any key " << endl;
//      waitBt();
    #endif
  }
  return ret;
}

int  Galerkin_Proj_KuraShi::dissOneStep(const interval x0[],interval h,
         interval xencl[],interval xstep[], int use_perturbation,
         const interval perturb_vect[])
   //  if use_perturbation then we add <perturb_vect> to the vector field of gal. proj.  KS
     // on input: \varphi([0,h],x0) \subset xencl
     // on output: in xstep - the bounds for \varphi(h,x)_i
     //           obtained from the dissipativity are returned
     //  these are bounds for Galerkin projection (without inclusion of the tail)
{
  interval *lperturb;
  lperturb=new interval[dim+1];
  copyvector(y_c,lperturb,dim);
  if(use_perturbation)
    for(int i=1;i<=dim;i++){
      lperturb[i]+=perturb_vect[i];
    }
  int ret=pkurashi->dissOneStep(x0,h,xencl,xstep,1,lperturb);
  delete [] lperturb;
  return ret;
}

int  Galerkin_Proj_KuraShi::incl_dissOneStep(const interval x0[],interval h,
              interval xencl[],interval xstep[])
  // on input: \varphi([0,h],x0) \subset xencl
     // On output: in xstep - the bounds for \varphi(h,x)_i
     //           obtained from the dissipativity are returned
     //  these are bounds for Galerkin projection + influence of the tail
{
  return pkurashi->dissOneStep(x0,h,xencl,xstep,1,inclusion_interval);
}


