#ifndef _multiint_h
#define _multiint_h


#define NORM_SUM 1
#define NORM_EUCL 2
#define NORM_MAX 3


void wrongdiminfo(int dim1,int dim2,const char procname[]);

char* norm_name(char* name,int maxlen, int norm_type);

interval norm(const interval x[],int pdim,int normfunc);
double norm(const double x[],int pdim,int normfunc);

interval norm(const interval m[][MAX_DIM+1],int pdim, int normfunc);
double norm(const double m[][MAX_DIM+1],int pdim, int normfunc);

interval normIVect(const interval & x,int pdim,int normfunc);
// returns the norm of  x^pdim=(x,x,...,x) - pdim-times repeated x


interval normMax(const interval x[],int pdim);
double normMax(const double x[],int pdim);

interval gradNormEucl(const interval x[],int pdim);
  //  computes eculidean norm of the 'gradient' i.e.
  //   vector of the form i*x[i], i=1..pdim

double gradNormEucl(const double x[],int pdim);
  //  computes euclidean norm of the 'gradient' i.e.
  //   vector of the form i*x[i], i=1..pdim


interval normEucl(const interval x[],int pdim);

double normEucl(const double x[],int pdim);

interval normSum(const interval x[],int pdim);
interval normSum(const interval m[][MAX_DIM+1],int pdim);
double normSum(const double m[][MAX_DIM+1],int pdim);

int vect_normalize(interval Vector[],int pdim,int normfunc);
// a normalization
// ON OUTPUT: 1-is returned if 0 \notin Vector  and Vector/=Norm(Vector)
//            0 - is returned if 0 \in Vector
int vect_normalize(double x[],int pdim,int normfunc);
// a normalization
// ON OUTPUT: 1-is returned if 0 \notin Vector  and Vector/=Norm(Vector)
//            0 - is returned if 0 \in Vector


interval normEucl(const interval m[][MAX_DIM+1],int pdim);
// returns an approximate euclidean operator norm of a matrix <m>
//   pdim - the dimension of <m>

double normEucl(const double m[][MAX_DIM+1],int pdim);
// returns an approximate euclidean operator norm of a matrix <m>
//   pdim - the dimension of <m>

interval normMax(const interval D[][MAX_DIM+1],int pdim);
double normMax(const double D[][MAX_DIM+1],int pdim);



interval lognormSum(const interval x[][MAX_DIM+1],int pdim);
// the logarithmic norm induced by the "sum norm" (L_1- norm) |x|=\sum |x_i|
// returns upper bound - an interval of zero-width

interval lognormMax(const interval x[][MAX_DIM+1],int pdim);
// the logarithmic norm induced by the "max norm" (L_\infty- norm) |x|=\max |x_i|
// returns upper bound - an interval of zero-width

interval logarithmic_norm(const interval x[][MAX_DIM+1],int pdim,int normfunc);
//  returns logarithmic norm of x (an upper bound);
//  when we use the euclidean norm - the procedure
//    is not rigorous; for max and sum - the function is rigorous

double logarithmic_norm(const double x[][MAX_DIM+1],int pdim,int normfunc);
//  returns logarithmic norm of x (an upper bound);





double distNormMax(const double x[][MAX_DIM+1],const double y[][MAX_DIM+1],
     int pdim);

double distNormMax(const double x[],const double y[],int pdim);

double distNormEucl(const double x[],const double y[],int pdim);





interval inf_increase_interval(const interval &inter);
//  increases the diameter of <inter> by a tiny amount


int thesameIbox(const interval m[],const interval n[],int dim);
// returns 1 when m[i]==n[i] for i=1,...,dim
//         0 otherwise

void MultiIhull(const interval m1[],const interval m2[],interval hull[],int pdim);
//  On output : hull=intervalhull(m1,m2);  pdim=dim(m1,m2,hull)

void MultiIhull(const interval m1[][MAX_DIM+1],const interval m2[][MAX_DIM+1],
        interval hull[][MAX_DIM+1],int pdim);
//  On output : hull=intervalhull(m1,m2);  pdim=dim(m1,m2,hull)


int MultiIn(const double m1[],const  interval m2[], int interior,int pdim);
// returns 1, when m1 \in m2
//          if interior we require m1 \subset interior(m2)
//  pdim - dimension

int MultiIn(const interval m1[],const interval m2[], int interior,int pdim);
// returns 1, when m1 \subset m2
//          if interior we require m1 \subset interior(m2)
//  pdim - dimension



 void mid(const interval s[],interval t[], int pdim);
//  t=mid(s)

 void mid(const interval s[][MAX_DIM+1],interval t[][MAX_DIM+1],int pdim);
//  t=mid(s)

double MultiIDiam(const interval m[],int pdim,int normfunc);
// computes the norm (defined by <normfunc>) of the vector build form
//  diameters of <m>

void MultiINeigh(const interval m[],interval diam,interval res[],int pdim);
// res = B_M(m,diam), a 'ball' of radius <diam> around m

void MultiISplit(interval m[],double *diam,int pdim,int normfunc);

void DecomposeIMulti(interval m[],interval r[],int pdim);
// on output m(output) + r(output) \supset m(input)
// r(output)[i]=[-a_i,a_i] - i.e. is symmetric
// diam(m(output))=0


interval smarteval_vAv(const interval c[],const interval a[][MAX_DIM+1],const interval b[],
      int dim);
// // computes c^T*a*b

void multAV(const double m[][MAX_DIM+1],const double x[], double r[],int pdim);
     // r=m*x

void multAV(const interval m[][MAX_DIM+1],const interval x[], interval r[],int pdim);
     // r=m*x

void multAV(const double m[][MAX_DIM+1],const interval x[], interval r[],int pdim);
     // r=m*x


void ArrArr(const interval A[][MAX_DIM+1],const interval B[][MAX_DIM+1],
   interval C[][MAX_DIM+1],int pdim);
  /* C=A*B */

void ArrArr(const double A[][MAX_DIM+1],const interval B[][MAX_DIM+1],
   interval C[][MAX_DIM+1],int pdim);
  /* C=A*B */

void ArrArr(const interval A[][MAX_DIM+1],const double B[][MAX_DIM+1],
          interval C[][MAX_DIM+1],int pdim);
  /* C=A*B */

void ArrArr(const double A[][MAX_DIM+1],const double B[][MAX_DIM+1],
   double C[][MAX_DIM+1],int pdim);
  /* C=A*B */

double arrfabs(const interval x[],int pdim);

void arrabs(const interval a[][MAX_DIM+1],double t[][MAX_DIM+1],int dim);
// on output:  t[i][j]=Right(abs(a[i][j]))


void transpose(const interval A[][MAX_DIM+1],interval C[][MAX_DIM+1],
   int pdim);
// C=A^t

void transpose(const double A[][MAX_DIM+1],double C[][MAX_DIM+1],
         int pdim);
// C=A^t

void transpose(const double A[][MAX_DIM+1],interval C[][MAX_DIM+1],
         int pdim);
// C=A^t

void copymatrix(const double S[][MAX_DIM+1],double T[][MAX_DIM+1],
         int pdim);
/* T=S */

void copymatrix(const double S[][MAX_DIM+1],interval T[][MAX_DIM+1],
         int pdim);
/* T=S */

void copymatrix(const interval S[][MAX_DIM+1],
    interval T[][MAX_DIM+1],int pdim);
/* T=S */


void copyvector(const double S[],double T[],int pdim);
// T=S

void copyvector(const double S[],interval T[],int pdim);
// T=S

void copyvector(const interval S[],interval T[],int pdim);
// T=S


/*
template<typename SourceType, typename TargetType>
void copyvector(const SourceType S[],TargetType T[],int pdim);
// T=S
*/

void copysection(const double S[],double T[],int pdim);
// the indices in section start from 0, hence we need different function
//   then <copyvector>
// T=S


void copysection(const interval S[],interval T[],int pdim);
// T=S



void identity(double m[][MAX_DIM+1],int dim);
// m=identity matrix

void identity(interval m[][MAX_DIM+1],int dim);
// m=identity matrix


void zero(interval m[][MAX_DIM+1],int dim);
// m=zero matrix

void zero(double m[][MAX_DIM+1],int dim);
// m=zero matrix

void zero(interval v[],int dim);
// v[i]=0, i=0,dim

void zero(double v[],int dim);
// v[i]=0, i=0,dim

void tointerval(const double x[],interval xi[],int dim);

void tointerval(const double x[][MAX_DIM+1],interval xi[][MAX_DIM+1], int dim);
//  xi=x

void todouble(const interval xi[],double x[],int dim);
// x=mid(xi)

void todouble(const interval xi[][MAX_DIM+1],double x[][MAX_DIM+1],
          int dim);


interval inflate(const interval &c,const interval &f);
// on output an interval <interout> is returned such that
//  center(interout)=center(c) and diam(interout)=f*diam(c)

void ArrInflate(interval t[],interval c, int dim);


bool commonpart(const interval & x1,const interval& x2,interval *cp);
// on output:  cp =intersection of <x1> and <x2>
//  returns true - when it is nonempty
//          false - otherwise, in this case cp - is not changed


bool commonpart(const interval x1[],const interval x2[],interval cp[],int dim);
// cp[i]=commonpart(x1[i],x2[i]), for i=1..dim
//  returns true - when it is nonempty
//          false - otherwise, in this case cp - is not changed

bool commonpart(const interval x1[][MAX_DIM+1],const interval x2[][MAX_DIM+1],
         interval cp[][MAX_DIM+1],int dim);
// cp[i][j]=commonpart(x1[i][j],x2[i][j]), for i,j=1..dim
//  returns true - when it is nonempty
//          false - otherwise, in this case cp - is not changed

void split(interval m[][MAX_DIM+1],interval r[][MAX_DIM+1],
             int dim);
//  m(input) \subset m(output)+r(output)
//   diam(m(output))=0

interval diameter(const interval m[][MAX_DIM+1],int dim);

interval diameter(const interval m[],int dim);


double det2x2(const double a[][MAX_DIM+1]);
//  dim a =2; det2x2=det(a)


double trace2x2(const double a[][MAX_DIM+1]);
// dim a = 2;   trace2x2=Trace (a)

int inverse(const double a[][MAX_DIM+1],double r[][MAX_DIM+1],int dim);
//  r=a^{-1} ;; dim must be 1 or 2



int d2invmatrix(const interval m[][MAX_DIM+1],interval inv[][MAX_DIM+1],int  dim );
// dim must be equal to 2


int diag2x2(const double df[][MAX_DIM+1],double *lambda1,double vec1[],
   double *lambda2,double vec2[]);
//  dim df=2
//  On output : (lambda1,vec1) - pair of eigenvalue and eigenvector
//              (lambda2,vec2) -   ''
// this procedure assumes that lambda1,lambda2 are real
//   and there is no nontrival Jordan cell




int diag2x2(const interval df[][MAX_DIM+1],interval *lambda1,interval vec1[],
   interval *lambda2,interval vec2[]);
//  dim df=2
//  On output : (lambda1,vec1) - pair of eigenvalue and eigenvector
//              (lambda2,vec2) -   ''
// the procedure assumes that lambda1,lamda2 are real
//   and there is no nontrival Jordan cell
// the function is not really finished, it does not treat the situation,
// when both off-diagonal terms contain 0


interval orthogonalcomponent(interval Vect[], interval Vnorm[],int pdim);
// on output <Vect> is equal to   Vect - <Vect|Vnorm>Vnorm
// i.e. we subtract from <Vect> its projection on <Vnorm>
//   function assumes that <VNorm> is normalized
//  we return <Vect|VNorm>


double orthogonalcomponent(double Vect[], double Vnorm[],int pdim);
// on output <Vect> is equal to   Vect - <Vect|Vnorm>Vnorm
// i.e. we subtract from <Vect> its projection on <Vnorm>
//   function assumes that <VNorm> is normalized
//  we return <Vect|VNorm>



interval scprod(const interval x[],const interval y[],int pdim);
double scprod(const double x[],const double y[],int pdim);
     /* calculates scalar product of two vectors*/


double evalsection(const double sec[],const double x[],int dim);
// the evaluation of the section form <sec> on <x>


interval evalsection(const interval sec[],const interval x[],int dim);
// the evaluation of the section form <sec> on <x>




void dropcoord(const double Dx[][MAX_DIM+1],int coord,
                        double Dp[][MAX_DIM+1],int dim);
  // from <dim> x <dim> matrix Dx the <coord> row and columns are
  // removed, result is placed in <Dp>
  // function used in computation of Jacobian of Poincare map


void dropcoord(const double D3in[][MAX_DIM+1][MAX_DIM+1],int coord,
               double  D3out[][MAX_DIM+1][MAX_DIM+1],int dim);
//  from dim x dim x dim matrix D3in
//  we delete all entries with the index i - and the resulting matrix is D3out


void dropcoord(const interval Dx[][MAX_DIM+1],int coord,
                        interval Dp[][MAX_DIM+1],int dim);
  // from <dim> x <dim> matrix Dx the <coord> row and columns are
  // removed, result is placed in <Dp>
  // function used in computation of Jacobian of Poincare map




void dropcoord(const double xin[],int coord,double xout[],int dim);
// From the <dim>-dimensional vector <xin> we drop <coord>-th coordinate
//  the resulting vector is stored in <xout>

void dropcoord(const interval xin[],int coord,interval xout[],int dim);
// From the <dim>-dimensional vector <xin> we drop <coord>-th coordinate
//  the resulting vector is stored in <xout>


void insertcoord(const interval xin[],int coord,interval xout[],interval valcoord,
        int dim);
//  We modify <dim-1>-dimensional vector <xin>, by inserting on <coord>'s slot
//  the value <valcoord>  the resulting vector <dim>-dimensional vector
//  is stored in <xout>

void insertcoord(const interval Din[][MAX_DIM+1],int coord,
                        interval Dout[][MAX_DIM+1],int dim);
  // In the <dim-1> x <dim-1> matrix <Din> we insert new coordinate
  //   this means new row (==0) and column (==0) -  <coord> position
  //  result is placed in <Dout>


int finddropcoord(const double sec[],int dim);
// function tries to test if sec is really of the form:
//   x_i = c_i ;
// in this case <i> is returned, otherwise 0


int finddropcoord(const interval sec[],int dim);
// function tries to test if sec is really of the form:
//   x_i = c_i ;
// in this case <i> is returned, otherwise 0

double zglrandom(double d,double u);
// function generates random number from the interval [d,u]


interval det(const interval m[][MAX_DIM+1],int dim);
//  returns the determinant by developing along the first row
//  det=\sum_{j=1}^dim (-1)^{i+j} det({\hat m}_{i,j})
//  where {\hat m}_{i,j} is the matrix obtained from <m> by removing
//  i-th row and j-th column

int kramers_inverse(const interval m[][MAX_DIM+1],interval minv[][MAX_DIM+1],int dim);
// on output m^{-1} \subset minv
// 1 - is returned if the inverse of <m> exists, <dim> - the dimension of <m>
//  Algorithm - the Kramers formula
//      m^{-1}_{ij}=(-1)^{i+j} det ({\hat m}_{j,i})/det(m)
//  where {\hat m}_{i,j} is the matrix obtained from <m> by removing
//  i-th row and j-th column


int inverse(const interval m[][MAX_DIM+1],interval minv[][MAX_DIM+1],int dim);
// on output m^{-1} \subset minv
// 1 - is returned if the inverse of <m> exists, <dim> - the dimension of <m>
//  Algorithm - Gauss elimination


int precond_inverse(const interval m[][MAX_DIM+1],interval minv[][MAX_DIM+1],int dim);
// on output m^{-1} \subset minv
// 1 - is returned if the inverse of <m> exists, <dim> - the dimension of <m>
// 0 - otherwise
//  Algorithm - the Gauss elimination; we use mid(minv) as an initial guess
//     to the inverse


int pos_defined(interval m[][MAX_DIM+1],int dim);
// m is dim x dim - dimensional matrix
// On output 1 - when m is positively defined
//           0 - otherwise



int intab(int k,const int tab[],int istart,int iend,int *indk);
// 1 - is returned if k=tab[j] for some istart<= j <= iend, on output*ind=j
// 0 otherwise

interval hausdorff_dist(interval u,interval v);
// returns the hausdorff distance between u and v


int gennewseqi(int seq[],const int l[], int pos, int dim);
//   the function generates the next sequence of integers in lexigraphic order
//  the matrixes <seq>, <l> have dimension <dim>
//  the range of integers in the i-th entry in matrix <seq> is
//   0,....,l[i]-1
//   pos - we are allowed to change  <pos,pos+1,...,dim> entries
//               in <seq>
//  The input sequence is in <seq>
//  1- is returned if we can generate a new sequence, which is then stored in <seq>
//  0 - otherwise
//////
//  the smallest seqence: seq[i]=0
//   the largest one seq[i]=l[i]-1


void divide_set(int dim,const int seq[],const int divisiontable[],
       int divdim,const interval ixb[],const interval C[][MAX_DIM+1],const interval r0[],
       interval pixb[],interval pr0[]);
 //  we  divide
 //  the set ixb + C*r0 into \Pi_{1<= i <=divdim}  divisiontable[i] parts
 //  the i-th direction in <r0> is divided into divisiontable[i] equal parts
 //  seq[i] for i=1,\dots,divdim , specifies the corresponding part
 //  seq[i]\in \{0,\dots,divisiontable[i]-1 \}
 //////////
 //  On output: the part corresponding to <seq> is given by
 //            pixb + C*pr0


void divide_set(int dim,const int seq[],const int divisiontable[],int divdim,
                              interval r0[],interval pr0[]);
 //  we  divide
 //  the set r0 into \Pi_{1<= i <=divdim}  divisiontable[i] parts
 //  the i-th direction in <r0> is divided into divisiontable[i] equal parts
 //  seq[i] for i=1,\dots,divdim , specifies the corresponding part
 //  seq[i]\in \{0,\dots,divisiontable[i]-1 \}
 //////////
 //  On output: the part corresponding to <seq> is given by pr0




bool orthogonalize(double q[][MAX_DIM+1],int pdim,int vectnum);
//  the function orthogonalizes <pdim>-dimensional vectors
//     q[1],q[2],\dots,q[vectnum]
//  On output:
//     q - contains orthonormalized vectors
//     1 - is returned if everything went ok
//     0 - otherwise (the vector of zero length was encountered during
//         orthogonalisation
//  the real number is considered to be effectively zero when it
//    is less than <smalldouble=1e-12>


bool orthogonalize(interval q[][MAX_DIM+1],int pdim,int vectnum);
//  the function orthogonalizes <pdim>-dimensional vectors
//     q[1],q[2],\dots,q[vectnum]
//  On output:
//     q - contains orthonormalized vectors
//     1 - is returned if everything went ok
//     0 - otherwise (the vector of zero length was encountered during
//         orthogonalisation


int solve_ax_r(interval a, interval *x, interval r);
// finds x, such that r \subset a* x
// we look for x with possibly small diameter


bool matrixexp(const interval a[][MAX_DIM+1],int dim,interval res[][MAX_DIM+1]);
//  this is an 'old' function, the 'new' one is <smart_matrixexp>
// on output: res=exp(a)
//  true - is returned when the computation was sucessful (convergence criterion
//           are met
//  false - otherwise

bool matrixexp(const double A[][MAX_DIM+1],const int dim,double expA[][MAX_DIM+1]);
// on output expA=exp(A)
//    the iteration is continued  till certain criteria are met
// if not then <false> is returned


bool smart_matrixexp(const double A[][MAX_DIM+1],const int dim,double expA[][MAX_DIM+1],
         int level=0);
// on output expA=exp(A)
//   if A - has too large norm, then exp(A) is computed by exp(A)=exp(A/2)*exp(A/2)
//       and exp(A/2)   might require another computation of expA/4) etc
//   this is a recursive function
//  level - indicates how deep in the recurrence are we

bool smart_matrixexp(const interval A[][MAX_DIM+1],const int dim,interval expA[][MAX_DIM+1],
         int level =0 );
// on output expA=exp(A)
//   if A - has too large norm, then exp(A) is computed by exp(A)=exp(A/2)*exp(A/2)
//       and exp(A/2)   might require another computation of expA/4) etc
//   this is a recursive function
//  level - indicates how deep in the recurrence are we




template <class numtype>
void increase_dim(numtype **s,int bdim,int newdim)
// On input: (*s) is bdim-dimensional vector, with indices starting from 1
// On output: if newdim > bdim
//         (*s) is newdim-dimensional vector, with indices starting from 1
{
  if(bdim >= newdim) return;

  numtype *t;
  t=new numtype[newdim+1];
  copyvector(*s,t,bdim);
  for(int i=bdim+1;i<=newdim;i++)
    t[i]=0;
  delete [] (*s);
  *s=t;
}

#endif  /* _multiint_h */

