#ifndef _oneshot_h_
#define _oneshot_h

#include"section.h"

///// ONE SHOT PROOF CONSTANTS
// what coordinates to use: for Rossler we can make the following choices:
//  1. standard - model of Dp - upper triangular
//  2. 1-eigenvector , 2 - orthogonal vector - model of DP - upper triangular
//                I expect 2 to be better than 1
//  3. eigenvectors - model of DP - diagonal, this should be difficult
//            numerically

// constants used to describe the above presented choices
  const int dpuppertriangular=1;
  const int dpdiagonal=2;



// constants describing our decisions on the location of the center
//  of our next try for P(N) \subset N
  const int middleofPN=1;   // take center of PN
  const int NPN_midlept=2;  // take center of intersection of N  and PN




int osa_coord(int sugested_coord);
// choice of the type of coordinates




void  write_no_inclusion(ostream &report,const interval fx[],const interval x[],
        int dim,const char *nfx,const char *nx);


void compute_ratios(const interval r0[],const interval fx[],int dim,int *numbigratios,
        double *largestratio);



void shift_mod_interval(int mustbeinx1, interval x1,interval x2,interval *xnew,double cmix);
// on output xnew =center(x2) + [-1,1]*d/2  (eventually intersected with x1)
//          where d <= diam(x1) and
//      if diam(x2) < diam(x1) then d=(1-cmix)diam(x2) + cmix diam(x1)
//   if mustbeinx1, then we additionally take care to make sure that xnew is contained in x1
/// In particular from the construction it follows that on output the following
//  condition is true:
//           if x2 \subset x1, then x2 \subset xnew


void inflate_mix_intervals(interval x,interval y,interval *xret,double cmix);
// we assume that y \subset x
// we inflate a bit <y> and return it in <xret> , on output y \subset xret \subset x
//  cmix \in [0,1)


int  estimate_dp(FPoincareMap &Pmap,double xb[],double delta[],
    interval intvDp[][MAX_DIM+1],int realonly);
// the function tries to estimate DP on the set xb + [-delta[i],delta[i]]
//  for the map <Pmap>
//  if <realonly> - double (nonrigorous) computations
//         otherwise we do first C1-computations, only if they look bad we
//       eventually perform nonrigorous ones


int rigorous_dp(FPoincareMap &Pmap,interval x[],interval Px[],interval intvDp[][MAX_DIM+1],
    estimates *estm, int useothercoord=0,LinCC *pcoordch=NULL);
// the function  estimates DP and P(x) rigorously on the set x for the map <Pmap>
//  is <useothercoord> - then x on input is expressed in <*pcoordch> coordinates
//  and on output <Px> and <intvDp> is also expressed in these coordinates


double lipschitzpartoferror(interval intvDp[][MAX_DIM+1],
   double dpmodel[][MAX_DIM+1],int dim);
// the function returns a rigorous bound in the max norm (L^\infty) for
//  the difference intvDp - dpmodel



int maponeshotattempt(FPoincareMap &Pmap,double xb[],double delta[],
            ofstream &report,int coord,int dimdiag);
//  checks Brouwer theorem for pmap - which can be any C1 map
//  xb - our approximate fixed point
//  delta[i] - starting tolerance for i-th coordinate, we will try to
//        perform proof inside these bounds, this serves also
//        as geometric indicator of shape when computing heuristic
//        constants
//  coord - possible values:
//         - standardcoord
//         - orthogonalizedeigenvectors;
//         - eigenvectors;
//         - partialeigenbasis
//         - eigen_subspaces_ortogonalized



void mapmanysizesattempt(int testfunc(FPoincareMap &,double[],double [],
   ofstream &,int,int), FPoincareMap &Pmap,double xb[],ofstream &report,
  double maxdelta,double mindelta,int coord,int dimdiag);
//  checks for various sizes that a map has a fixed point (testfunc)
//    PMap  - C1 map
//  xb - our approximate fixed point
//  maxdelta  and mindelta - bounds on the size of the sides of N
//  coord - possible values:
//         - standardcoord
//         - orthogonalizedeigenvectors;
//         - eigenvectors;
//         - partialeigenbasis
//         - eigen_subspaces_ortogonalized
//  parameters of testfunc: PMap, xcenter, delta[], report, coord, dimdiag


int C1computefixP(FPoincareMap &Pmap,double xb[],double delta[],
            ofstream &report,int coord,int dimdiag);
//  the function computes P(xb + delta) and dP(xb) - rigorously
//  xb - our approximate fixed point
//  delta[i] - starting tolerance for i-th coordinate
//  coord - possible values:
//         - standardcoord
//         - orthogonalizedeigenvectors;
//         - eigenvectors;
//         - partialeigenbasis
//         - eigen_subspaces_ortogonalized







int simpleintervaliteration(FPoincareMap &PMap,double xb[],double delta[],
       ofstream &report,int coord,int dimdiag,int iternum,int divtable[],int dimdiv);
// xb - a good approximation the attracting fixed point
// iternum - the number of iterates
// this function takes the set N=xb + \\Pi_i[-delta[i],delta[i]] (in coordinates
//               specified by by <coord>)
//  if dimdiv >0 , then N_i is divided into divtable[i] equal parts
//               hence the set N is divided into \Pi_{1<= i<=divdim} divtable[i]}
// it computes P(N) ; for the next iterate N=P(N) \\cap N and so on
//// returns 1 - when PROVED P(N) \subset N



void testdiagfixedp(ostream &report,FMap *PMap,int mapiter,
                                  double xb[],int dimdiag);


#endif
